#!/usr/bin/env python

# Neighborhood Correlation common routines
# (see http://www.neighborhoodcorrelation.org)

# (C) 2008 Jacob Joseph <jmjoseph@andrew.cmu.edu>,
#          Nan Song, and Carnegie Mellon University
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import sys
from math import log, log10, sqrt

class nc_base( object):
    """Common Neighborhood Correlation routines"""

    blast_info = None          # dict of BLAST run parameters
    nc_info = None             # dict of NC run parameters    
    
    all_seqs = None            # List of all sequences
    score_dict = None          # BLAST score dictionary

    var_cache = None           # Sequence variance (var_x)
    E_cache = None             # (n_x, sum_x, E_x) for each sequence

    def __init__(self):

        self.blast_info = {}
        self.nc_info = {}

        # Use python implementations of (co)variance and expectation
        self.var_xx = self.var_xx_python
        self.E_x = self.E_x_python
        self.var_xy = self.var_xy_python

        self.var_cache = {}
        self.E_cache = {}

    def new_run( self, br_id, e_thresh, nc_thresh, smin_factor,
                 self_hits, num_residues=None):
        
        self.nc_info = { 'br_id' : br_id,
                         'e_thresh' : e_thresh,
                         'nc_thresh' : nc_thresh,
                         'smin_factor' : smin_factor,
                         'score_type' : 'bit_score',
                         'self_hits' : self_hits}

        if num_residues is not None:
            self.blast_info['num_residues'] = num_residues

        
    def calculate_smin( self):

        assert len(self.blast_info)>0, """blast_info not initialized.  fetch_score_dict_flatfile must first be
executed, or a database call made"""

        # Fetch blast run informatino
        num_sequences = self.blast_info['num_sequences']
        num_residues = self.blast_info['num_residues']

        blast_e_param = self.blast_info['params']['expectation']

        smin_factor = self.nc_info['smin_factor']

        smin = (log( (num_residues ** 2) / blast_e_param)
                / log(2))

        # reduce the true smin by some factor
        smin = smin * smin_factor
        self.nc_info['smin'] = smin


    def fetch_score_dict_flatfile( self, fname):
        print >> sys.stderr, "Reading Score File" 

        fd = open( fname)
        self.score_dict = {}
        sd = self.score_dict

        # Use the maximum score for bit_score, otherwise, use the
        # minimum, which would be appropriate for e_value
        if self.nc_info['score_type'] == 'bit_score':
            comp_fn = lambda a,b: a > b
        else:
            comp_fn = lambda a,b: a < b
        

        # sequence id map.  Used to avoid repeatedly storing keys
        # objects in memory.
        seq_id_map = {}
        
        for l in fd:
            larr = l.strip().split()
            assert len(larr)==3, "Flatfile line unparsable: '%s'" % l

            (seq_id_0, seq_id_1, score) = larr
            
            # use log10(bit_score)
            score = log10( float(score))

            if not seq_id_0 in seq_id_map: seq_id_map[seq_id_0] = seq_id_0
            else: seq_id_0 = seq_id_map[seq_id_0]
            
            if not seq_id_1 in seq_id_map: seq_id_map[seq_id_1] = seq_id_1
            else: seq_id_1 = seq_id_map[seq_id_1]

            # make score symmetric
            if not seq_id_0 in sd: sd[seq_id_0] = {}
            if not seq_id_1 in sd: sd[seq_id_1] = {}

            rev_score = sd[seq_id_1].get( seq_id_0, None)

            if rev_score is not None and comp_fn( rev_score, score):
                sd[seq_id_0][seq_id_1] = rev_score
            else:
                sd[seq_id_0][seq_id_1] = score
                sd[seq_id_1][seq_id_0] = score

        fd.close()
        self.all_seqs = seq_id_map.keys()

        # provide information about the blast run. Used to calculate smin
        self.blast_info['num_sequences'] = len(self.all_seqs)
        if not 'num_residues' in self.blast_info:
            self.blast_info['num_residues'] = int(537.2148 * len(self.all_seqs))
        self.blast_info['params'] = {'expectation': 10 * len(self.all_seqs)}
        return


    def var_xx_python( self, seq_id):
        "Calculate and cache the variance of one sequence dictionary"

        if self.var_cache is None:
            self.var_cache = {}

        # return if already calculated
        if self.var_cache.has_key( seq_id):
            return self.var_cache[seq_id]

        # Fetch the mean
        if self.E_cache.has_key( seq_id):
            (n_0, sum_0, mean) = self.E_cache[ seq_id]
        else:
            (n_0, sum_0, mean) = self.E_x(seq_id)

        # look these up once
        if not 'smin' in self.nc_info: self.calculate_smin()
            
        lsmin = log10( self.nc_info['smin'])
        num_sequences = self.blast_info['num_sequences']

        score_list = self.score_dict.get(seq_id, {}).values()
        others = num_sequences - n_0

        # now compute the variance, first with real values, then
        # adding all others
        sum2 = 0
        for score in score_list:
            sum2 += (score - mean)**2
        sum2 += ((lsmin - mean)**2) * others

        var = sum2 / (num_sequences - 1)
        
        # Precision issues might lead to variances below zero
        # Variance is set to zero if no neighbors
        if var < 0 or n_0==0:
            #print "zeroed variance for:", seq_id, var
            var = 0

        self.var_cache[seq_id] = var
        return var


    def E_x_python( self, seq_id):
        "Calculate and cache n_x, sum_x, and E_x"

        if not 'smin' in self.nc_info: self.calculate_smin()
        logsmin = log10( self.nc_info['smin'])
        num_sequences = self.blast_info['num_sequences']

        if self.E_cache is None:
            self.E_cache = {}
        
        if self.E_cache.has_key(seq_id):
            return self.E_cache[seq_id]

        vals = self.score_dict.get(seq_id, {}).values()
        n_x = len(vals)
        others = num_sequences - n_x
        sum_x = sum(vals)
        E_x = ( sum_x + logsmin * others) / num_sequences

        self.E_cache[ seq_id] = (n_x, sum_x, E_x)
        return (n_x, sum_x, E_x)


    def var_xy_python( self, seq_id_0, seq_id_1):
        "Calculate the correlation coefficient for two sequence dictionaries"

        # look these up once
        if not 'smin' in self.nc_info: self.calculate_smin()
        logsmin = log10(self.nc_info['smin'])
        num_sequences = self.blast_info['num_sequences']
        score_dict = self.score_dict
        E_cache = self.E_cache

        try: score_subdict_0 = score_dict[seq_id_0]
        except: score_subdict_0 = {}
        try: score_subdict_1 = score_dict[seq_id_1]
        except: score_subdict_1 = {}

        # grab Ex, Ey
        (n_0, sum_0, Ex) = E_cache[seq_id_0]
        (n_1, sum_1, Ey) = E_cache[seq_id_1]

        keys0 = score_subdict_0.keys()
        keys1 = score_subdict_1.keys()
        k_01 = set(keys0).intersection(keys1)  # this is a set, not a list
        n_01 = len(k_01)

        s01_prodsum = 0  # product of xy
        s01_sum = 0
        for k in k_01:
            x = score_subdict_0[k]
            y = score_subdict_1[k]
            s01_prodsum += x * y
            s01_sum += x + y

        others_union = num_sequences - (n_0 + n_1 - n_01)
        
        Exy = (s01_prodsum + (logsmin**2) * others_union +
               (sum_0 + sum_1 - s01_sum) * logsmin
               ) / num_sequences

        var_xy =  Exy - Ex * Ey

        return (n_0, n_1, n_01, var_xy)


    def calculate_nc( self, seq_id_0, seq_id_1):
        "Calculate the NC score for a single pair"

        var_0 = self.var_cache[seq_id_0]
        var_1 = self.var_cache[seq_id_1]

        if var_0 == 0 or var_1 == 0:
            #print "zero variance:", seq_id_0, seq_id_1, var_0, var_1
            #print "This may occur if a sequence has no matches at all."
            return None
        
        # equivalent to n_x, n_y, n_xy
        (n_0, n_1, n_01, var_01) = self.var_xy( seq_id_0, seq_id_1)
        nc = var_01 / sqrt( var_0 * var_1)

        return (n_0, n_1, n_01, nc)
    

    def calculate_nc_queries( self, query_seqs=None, out_fd=None):
        """Calculate all pairs of NC scores, reporting those above the
        nc_thresh.  Output is printed to stdout"""


        if not 'smin' in self.nc_info: self.calculate_smin()
        print >> sys.stderr, "  num_sequences:", self.blast_info['num_sequences']
        print >> sys.stderr, "   num_residues:", self.blast_info['num_residues']
        print >> sys.stderr, "           smin:", self.nc_info['smin']

        if out_fd is None: out_fd = sys.stdout

        if query_seqs is None: query_seqs = self.all_seqs

        # Fetch dictionary of all symmetric blast hits.
        score_dict = self.score_dict
        
        # cache the variance for each sequence now
        print >> sys.stderr, "Calculating Variances"
        for seq_id in self.all_seqs:
            self.var_xx(seq_id)

        nc_thresh = self.nc_info['nc_thresh']

        print >> sys.stderr, "Calculating NC Scores"
        for seq_id_0 in query_seqs:
            for seq_id_1 in self.all_seqs:

                retval = self.calculate_nc( seq_id_0, seq_id_1)
                (n_0, n_1, n_01, nc) = retval

                # only consider if above the threshold and no blast hit
                if nc < nc_thresh and not score_dict.get(seq_id_0,{}).has_key(seq_id_1):
                    continue

                print >> out_fd, seq_id_0, seq_id_1, nc #, n_0, n_1, n_01



